# encoding: utf-8

# Exemplo de implementação de lista encadeada simples, arranjo dinâmico,
# fila e pilha.
#
# Página: http://malbarbo.pro.br/ensino/2013/7256/
# Autor : Marco A L Barbosa <marco@malbarbo.pro.br>

from aed import Struct, executa_testes

__all__ = ['ListaEncadeada', 'ArranjoDinamico', 'Fila', 'Pilha']

##############################
# Lista encadeada

class No(Struct):
    '''
    Representa um Nó em uma lista encadeada,
        valor : Qualquer - é o valor armazenado no nó
        prox : No - é uma referência para o próximo nó
    '''
    def __init__(self, valor, prox):
        self.init(valor, prox)


class ListaEncadeada(Struct):
    '''
    Representa uma lista encadeada.
        _primeiro : No - é uma referência para o primeiro nó da lista
        _ultimo : No - é uma referência para o último nó da lista
    '''
    def __init__(self):
        self.init()
        self._primeiro = None
        self._ultimo = None

    def insere_inicio(self, v):
        '''
        ListaEncadeada, Qualquer -> None
        Insere v no início da lista.
        Exemplos:
        >>> xs = ListaEncadeada()
        >>> xs.insere_inicio(3)
        >>> xs._primeiro
        No(3, None)
        >>> xs._ultimo
        No(3, None)
        >>> xs.insere_inicio(1)
        >>> xs._primeiro
        No(1, No(3, None))
        >>> xs._ultimo
        No(3, None)
        >>> xs.insere_inicio(27)
        >>> xs._primeiro
        No(27, No(1, No(3, None)))
        >>> xs._ultimo
        No(3, None)
        '''
        self._primeiro = No(v, self._primeiro)
        if self._ultimo == None:
            self._ultimo = self._primeiro

    def insere_fim(self, v):
        '''
        ListaEncadeada, Qualquer -> None
        Insere v no final da lista.
        Exemplos:
        >>> xs = ListaEncadeada()
        >>> xs.insere_fim(3)
        >>> xs._primeiro
        No(3, None)
        >>> xs._ultimo
        No(3, None)
        >>> xs.insere_fim(10)
        >>> xs._primeiro
        No(3, No(10, None))
        >>> xs._ultimo
        No(10, None)
        >>> xs.insere_fim(20)
        >>> xs._primeiro
        No(3, No(10, No(20, None)))
        >>> xs._ultimo
        No(20, None)
        '''
        p = No(v, None)
        if self._primeiro == None:
            self._primeiro = p
            self._ultimo = p
        else:
            self._ultimo.prox = p
            self._ultimo = p

    def insere(self, pos, v):
        '''
        ListaEncadeada, Natural, Qualquer -> None
        Insere v na posição pos.
        Exemplos:
        >>> xs = ListaEncadeada()
        >>> xs.insere(0, 4)
        >>> xs._primeiro
        No(4, None)
        >>> xs._ultimo
        No(4, None)
        >>> xs.insere(0, 8)
        >>> xs._primeiro
        No(8, No(4, None))
        >>> xs._ultimo
        No(4, None)
        >>> xs.insere(1, 6)
        >>> xs._primeiro
        No(8, No(6, No(4, None)))
        >>> xs._ultimo
        No(4, None)
        >>> xs.insere(3, 2)
        >>> xs._primeiro
        No(8, No(6, No(4, No(2, None))))
        >>> xs._ultimo
        No(2, None)
        >>> xs.insere(10, 20)
        Traceback (most recent call last):
            ...
        Exception: índice fora do intervalo
        '''
        if pos == 0:
            self.insere_inicio(v)
        else:
            # procura o nó na posição anterior a inserção
            p = self._consulta_pos(pos - 1)
            # insere após o nó encontrado
            self._insere_prox(p, v)

    def _insere_prox(self, p, v):
        '''
        ListaEncadeada, No, Qualquer -> None
        Insere o v após o nó p.
        Exemplos: veja o método insere
        '''
        p.prox = No(v, p.prox)
        if p.prox.prox == None:
            self._ultimo = p.prox

    def remove_inicio(self):
        '''
        ListaEncadeada -> Qualquer
        Remove o primeiro elemento da lista e devolve o valor removido.
        Lança uma exceção se a lista estiver vazia.
        >>> xs = ListaEncadeada()
        >>> xs.remove_inicio()
        Traceback (most recent call last):
            ...
        Exception: lista vazia
        >>> xs.insere_inicio(3)
        >>> xs.insere_inicio(5)
        >>> xs.remove_inicio()
        5
        >>> xs._primeiro
        No(3, None)
        >>> xs._ultimo
        No(3, None)
        >>> xs.remove_inicio()
        3
        >>> xs._primeiro == None
        True
        >>> xs._ultimo == None
        True
        '''
        if self._primeiro == None:
            raise Exception('lista vazia')
        v = self._primeiro.valor
        self._primeiro = self._primeiro.prox
        if self._primeiro == None:
            self._ultimo = None
        return v

    def remove_fim(self):
        '''
        ListaEncadeada -> None
        Remove o último elemento da lista e devolve o valor removido.
        Lança uma exceção se a lista estiver vazia.
        >>> xs = ListaEncadeada()
        >>> xs.remove_fim()
        Traceback (most recent call last):
            ...
        Exception: lista vazia
        >>> xs.insere_fim(3)
        >>> xs.insere_fim(5)
        >>> xs.insere_fim(8)
        >>> xs.remove_fim()
        8
        >>> xs._primeiro
        No(3, No(5, None))
        >>> xs._ultimo
        No(5, None)
        >>> xs.remove_fim()
        5
        >>> xs._primeiro
        No(3, None)
        >>> xs._ultimo
        No(3, None)
        >>> xs.remove_fim()
        3
        >>> xs._primeiro == None
        True
        >>> xs._ultimo == None
        True
        '''        
        if self._primeiro == None:
            raise Exception('lista vazia')
        if self._primeiro == self._ultimo:
            return self.remove_inicio()
        # procura o penúltimo nó
        p = self._primeiro
        while p.prox != self._ultimo:
            p = p.prox
        return self._remove_prox(p)

    def _remove_prox(self, p):
        '''
        ListaEncadeada, No -> Qualquer
        Remove o próximo a partir de p e devolve o valor armazenado no nó
        removido.
        '''
        v = p.prox.valor
        p.prox = p.prox.prox
        if p.prox == None:
            self._ultimo = p
        return v

    def remove(self, v):
        '''
        ListaEncadeada, Qualquer -> None
        Remove a primeira ocorrência de v da lista.
        Lança um exceção se x não está na lista.
        Exemplos
        >>> xs = ListaEncadeada()
        >>> xs.insere_fim(3)
        >>> xs.insere_fim(5)
        >>> xs.insere_fim(7)
        >>> xs.insere_fim(9)
        >>> xs.remove(20)
        Traceback (most recent call last):
            ...
        Exception: 20 não encontrado
        >>> xs.remove(3)
        >>> xs._primeiro
        No(5, No(7, No(9, None)))
        >>> xs._ultimo
        No(9, None)
        >>> xs.remove(7)
        >>> xs._primeiro
        No(5, No(9, None))
        >>> xs._ultimo
        No(9, None)
        >>> xs.insere_inicio(10)
        >>> xs.remove(9)
        >>> xs._primeiro
        No(10, No(5, None))
        >>> xs._ultimo
        No(5, None)
        >>> xs.remove(10)
        >>> xs._primeiro
        No(5, None)
        >>> xs._ultimo
        No(5, None)
        >>> xs.remove(5)
        >>> xs._primeiro == None
        True
        >>> xs._ultimo == None
        True
        '''
        if self._primeiro == None:
            raise Exception('lista vazia')
        if self._primeiro.valor == v:
            self.remove_inicio()
            return
        p = self._primeiro
        while p.prox != self._ultimo and p.prox.valor != v:
            p = p.prox
        if p.prox.valor != v:
            raise Exception(str(v) + ' não encontrado')
        self._remove_prox(p)

    def remove_pos(self, pos):
        '''
        ListaEncadeada, Natural -> Qualquer
        >>> xs = ListaEncadeada()
        >>> xs.insere_fim(5)
        >>> xs.insere_fim(2)
        >>> xs.insere_fim(4)
        >>> xs.insere_fim(8)
        >>> xs.remove_pos(2)
        4
        >>> xs.remove_pos(0)
        5
        >>> xs.remove_pos(1)
        8
        '''
        if pos == 0:
            return self.remove_inicio()
        p = self._consulta_pos(pos - 1)
        return self._remove_prox(p)

    def _consulta_pos(self, pos):
        '''
        ListaEncadeada, Natural -> Qualquer
        Devolve o No na posição pos.
        Lança uma exceção se pos >= len(lista) ou pos < 0.
        TODO: tratar índices negativos
        Veja os exemplos de __getitem__ e __setitem__
        '''
        p = self._primeiro
        i = pos
        while p != None and i > 0:
            p = p.prox
            i = i - 1
        if p != None and i == 0:
            return p
        else:
            raise Exception('índice fora do intervalo')


    def __getitem__(self, pos):
        '''
        ListaEncadeada, Natural -> Qualquer
        Devolve o elemento na posição i da lista.
        Lança uma exceção se pos >= len(lista) ou pos < 0.
        TODO: tratar índices negativos
        Exemplos:
        >>> xs = ListaEncadeada()
        >>> xs.insere_fim(7)
        >>> xs.insere_fim(2)
        >>> xs.insere_fim(10)
        >>> xs[0]
        7
        >>> xs[1]
        2
        >>> xs[2]
        10
        >>> xs[3]
        Traceback (most recent call last):
            ...
        Exception: índice fora do intervalo
        '''
        p = self._consulta_pos(pos)
        return p.valor

    def __setitem__(self, pos, v):
        '''
        ListaEncadeada, Natural, Qualquer -> None
        Modifica o valor na posição i para v.
        Lança uma exceção se pos >= len(lista) ou pos < 0.
        TODO: tratar índices negativos
        Exemplos:
        >>> xs = ListaEncadeada()
        >>> xs.insere_fim(7)
        >>> xs.insere_fim(2)
        >>> xs.insere_fim(10)
        >>> xs[0] = 40
        >>> xs[1] = 50
        >>> xs[2] = 60
        >>> xs
        ListaEncadeada([40, 50, 60])
        >>> xs[3] = 70
        Traceback (most recent call last):
            ...
        Exception: índice fora do intervalo
        '''
        p = self._consulta_pos(pos)
        p.valor = v

    def __in__(self, v):
        '''
        ListaEncadeada, Qualquer -> Boolean
        Devolve True se v está na lista, False caso contrário
        Exemplos:        
        >>> xs = ListaEncadeada()
        >>> xs.insere_fim(4)
        >>> xs.insere_fim(2)
        >>> xs.insere_fim(6)
        >>> 4 in xs
        True
        >>> 2 in xs
        True
        >>> 6 in xs
        True
        >>> 5 in xs
        False
        '''
        p = self._primeiro
        while p != None and p.valor != v:
            p = p.prox
        return p != None and p.valor == v

    def __len__(self):
        '''
        ListaEncadeada -> Natural
        Devolve o tamanho da lista, isto é, a quantidade de itens.
        Exemplos:        
        >>> xs = ListaEncadeada()
        >>> len(xs)
        0
        >>> xs.insere_inicio(4)
        >>> len(xs)
        1
        >>> xs.insere_inicio(8)
        >>> xs.insere_inicio(18)
        >>> len(xs)
        3
        '''
        tamanho = 0
        p = self._primeiro        
        while p != None:
            p = p.prox
            tamanho = tamanho + 1
        return tamanho

    def __iter__(self):
        '''
        ListaEncadeada -> Iterador
        Devolve um iterador para os elementos da lista.
        Exemplos:
        >>> xs = ListaEncadeada()
        >>> xs.insere_fim(3)
        >>> xs.insere_fim(7)
        >>> xs.insere_fim(10)
        >>> r = []
        >>> for x in xs:
        ...     r.append(x)
        >>> r
        [3, 7, 10]
        '''
        p = self._primeiro
        while p != None:
            yield p.valor
            p = p.prox

    def __repr__(self, nome='ListaEncadeada'):
        '''
        ListaEncadeada -> String
        Devolve uma representação textual da lista.
        Exemplos:
        >>> xs = ListaEncadeada()
        >>> xs
        ListaEncadeada([])
        >>> xs.insere_inicio(8)
        >>> xs
        ListaEncadeada([8])
        >>> xs.insere_fim(10)
        >>> xs
        ListaEncadeada([8, 10])
        >>> xs.insere_inicio(4)
        >>> xs
        ListaEncadeada([4, 8, 10])
        '''
        p = self._primeiro
        s = nome + '(['
        if p != None:
            s = s + str(p.valor)
            p = p.prox
        while p != None:
            s = s + ', ' + str(p.valor)
            p = p.prox
        s = s + '])'
        return s



##############################
# Arranjo dinâmico

class ArranjoDinamico(Struct):
    '''
    Representa um arranjo dinâmico, isto é, um arranjo que pode aumentar ou
    diminuir de tamanho. Observe que os arranjos em Python são dinâmicos, desta
    forma, esta classe não é necessária. O propósito desta classe é mostrar
    como um arranjo dinâmico pode ser implementado em termos de um arranjo
    estático.
        _valores : Arranjo - é o arranjo estático que armazena os valores
        _tamanho : Natural - é a quantidade de posições que de fatos estão
                             sendo usadas no arranjo _valores
    '''
    def __init__(self):
        self.init()
        self._valores = [None] * 10
        self._tamanho = 0

    def insere_inicio(self, v):
        '''
        ArranjoDinamico, Qualquer -> None
        Insere v no início do arranjo.
        Exemplos:
        >>> xs = ArranjoDinamico()
        >>> xs.insere_inicio(1)
        >>> xs
        ArranjoDinamico([1])
        >>> xs.insere_inicio(5)
        >>> xs
        ArranjoDinamico([5, 1])
        >>> for x in range(0, 10):
        ...     xs.insere_inicio(x)
        >>> xs
        ArranjoDinamico([9, 8, 7, 6, 5, 4, 3, 2, 1, 0, 5, 1])
        '''
        self.insere(0, v)

    def insere_fim(self, v):
        '''
        ArranjoDinamico, Qualquer -> None
        Insere v no fim do arranjo.
        Exemplos:
        >>> xs = ArranjoDinamico()
        >>> xs.insere_fim(1)
        >>> xs
        ArranjoDinamico([1])
        >>> xs.insere_fim(5)
        >>> xs
        ArranjoDinamico([1, 5])
        >>> for x in range(0, 10):
        ...     xs.insere_fim(x)
        >>> xs
        ArranjoDinamico([1, 5, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9])
        '''
        self.insere(self._tamanho, v)

    def insere(self, pos, v):
        '''
        ArranjoDinamico, Natural, Qualquer -> None
        Insere v na posição pos.
             Exemplos:
        >>> xs = ArranjoDinamico()
        >>> xs.insere(0, 4)
        >>> xs
        ArranjoDinamico([4])
        >>> xs.insere(0, 8)
        >>> xs
        ArranjoDinamico([8, 4])
        >>> xs.insere(1, 6)
        >>> xs
        ArranjoDinamico([8, 6, 4])
        >>> xs.insere(3, 2)
        >>> xs
        ArranjoDinamico([8, 6, 4, 2])
        >>> xs.insere(10, 20)
        Traceback (most recent call last):
            ...
        Exception: índice fora do intervalo
        '''
        if pos > self._tamanho or pos < 0:
            raise Exception('índice fora do intervalo')
        if self._tamanho == len(self._valores):
            # dobra de tamanho
            valores = [None] * len(self._valores) * 2
            for i in range(0, len(self._valores)):
                valores[i] = self._valores[i]
            self._valores = valores
        i = self._tamanho
        while i > pos:
            self._valores[i] = self._valores[i-1]
            i = i - 1
        self._valores[i] = v
        self._tamanho = self._tamanho + 1

    def remove_inicio(self):
        '''
        ArranjoDinamico -> Qualquer
        Remove o elemento do início do arranjo.
        Exemplos:
        >>> xs = ArranjoDinamico()
        >>> xs.insere_fim(5)
        >>> xs.insere_fim(3)
        >>> xs.insere_fim(7)
        >>> xs.remove_inicio()
        5
        >>> xs
        ArranjoDinamico([3, 7])
        '''
        return self.remove_pos(0)

    def remove_fim(self):
        '''
        ArranjoDinamico -> Qualquer
        Remove o elemento do fim do arranjo.
        Exemplos:
        >>> xs = ArranjoDinamico()
        >>> xs.insere_fim(5)
        >>> xs.insere_fim(3)
        >>> xs.insere_fim(7)
        >>> xs.remove_fim()
        7
        >>> xs
        ArranjoDinamico([5, 3])
        '''
        return self.remove_pos(self._tamanho - 1)

    def remove(self, v):
        '''
        ArranjoDinamico -> Qualquer
        Remove v do arranjo.
        Lança uma exceção se v não está no arranjo.
        Exemplos:
        >>> xs = ArranjoDinamico()
        >>> xs.insere_fim(5)
        >>> xs.insere_fim(3)
        >>> xs.insere_fim(7)
        >>> xs.remove(3)
        >>> xs
        ArranjoDinamico([5, 7])
        >>> xs.remove(20)
        Traceback (most recent call last):
            ...
        Exception: 20 não encontrado
        '''
        i = 0
        while i < self._tamanho and self._valores[i] != v:
            i = i + 1
        if i < self._tamanho and self._valores[i] == v:
            self.remove_pos(i)
        else:
            raise Exception(str(v) + ' não encontrado')

    def remove_pos(self, pos):
        '''
        ArranjoDinamico, Natural -> None
        Remove o elemento da posição pos.
        Lança uma exceção se pos < 0 ou pos >= len(self)
        Exemplos:
        >>> xs = ArranjoDinamico()
        >>> xs.insere_fim(5)
        >>> xs.insere_fim(7)
        >>> xs.insere_fim(2)
        >>> xs.insere_fim(3)
        >>> xs.remove_pos(4)
        Traceback (most recent call last):
            ...
        Exception: índice fora do intervalo
        >>> xs.remove_pos(0)
        5
        >>> xs
        ArranjoDinamico([7, 2, 3])
        >>> xs.remove_pos(1)
        2
        >>> xs
        ArranjoDinamico([7, 3])
        >>> xs.remove_pos(1)
        3
        >>> xs
        ArranjoDinamico([7])
        >>> xs.remove_pos(0)
        7
        >>> xs
        ArranjoDinamico([])
        '''
        self._verifica_pos(pos)
        v = self._valores[pos]
        for i in range(pos, self._tamanho - 1):
            self._valores[i] = self._valores[i + 1]        
        self._valores[self._tamanho] = None
        self._tamanho = self._tamanho - 1
        return v

    def _verifica_pos(self, pos):
        if pos >= self._tamanho or pos < 0:
            raise Exception('índice fora do intervalo')
        
    def __getitem__(self, pos):
        self._verifica_pos(pos)
        return self._valores[pos]

    def __setitem__(self, pos, v):
        self._verifica_pos(pos)
        self._valores[pos] = v

    def __len__(self):
        return self._tamanho

    def __iter__(self):
        for i in range(0, self._tamanho):
            yield self._valores[i]

    def __repr__(self, nome='ArranjoDinamico'):
        s = nome + '(['
        if len(self) >= 1:
            s = s + str(self[0])
        for i in range(1, len(self)):
            s = s + ', ' + str(self[i])
        s = s + '])'
        return s


##############################
# Fila

class Fila(Struct):
    '''
    Representa uma fila. Implementada usando lista encadeada.
    Exemplos:
    >>> f = Fila()
    >>> f.insere(3)
    >>> f.insere(5)
    >>> f.insere(8)
    >>> f.primeiro()
    3
    >>> f.remove()
    3
    >>> f.insere(10)
    >>> f.insere(20)
    >>> f.primeiro()
    5
    >>> f.remove()
    5
    >>> f.remove()
    8
    >>> f.remove()
    10
    >>> f.remove()
    20
    '''
    def __init__(self):
        self.init()
        self._lista = ListaEncadeada()

    def primeiro(self):
        return self._lista[0]

    def insere(self, v):
        self._lista.insere_fim(v)

    def remove(self):
        return self._lista.remove_inicio()

    def __repr__(self):
        return self._lista.__repr__('Fila')


##############################
# Pilha

class Pilha(Struct):
    '''
    Representa um pilha. Implementada usando lista encadeada.
    Exemplos:
    >>> p = Pilha()
    >>> p.insere(3)
    >>> p.insere(5)
    >>> p.insere(8)
    >>> p.topo()
    8
    >>> p.remove()
    8
    >>> p.insere(10)
    >>> p.insere(20)
    >>> p.topo()
    20
    >>> p.remove()
    20
    >>> p.remove()
    10
    >>> p.remove()
    5
    >>> p.remove()
    3
    '''
    def __init__(self):
        self.init()
        self._lista = ListaEncadeada()

    def topo(self):
        return self._lista[0]

    def insere(self, v):
        self._lista.insere_inicio(v)

    def remove(self):
        return self._lista.remove_inicio()
    
    def __repr__(self):
        return self._lista.__repr__('Pilha')


if __name__ == '__main__':
    executa_testes()
